#!/usr/bin/env python3
# coding=utf-8
# SPDX-License-Identifier: MIT
"""Contains the basic parent provider that all functional providers subclass and inherit from."""
from __future__ import annotations

import abc
import collections.abc
import os

import mcfonts.constants
import mcfonts.glyphs


class BaseProvider(abc.ABC):
    """The base provider class. All providers subclass this."""

    minimum_pack_format: int = 0
    # This means nothing, but it's fun.
    icon: str = "?"

    def __init__(
        self,
        provider_type: str,
        contents: dict,
        origin: os.PathLike[str] | str,
        chars_covered: set[str] | None = None,
    ):
        """
        :param provider_type: A string of the value in the JSON "type" field.
        :param contents: The provider's contents.
        :param origin: The absolute path to the provider's JSON file.
        :param chars_covered: A set of the individual characters covered by this provider.
        """
        if chars_covered is None:
            chars_covered = set()
        self.provider_type = provider_type
        self.chars_covered: set[str] = chars_covered
        self.contents: dict = contents
        self.origin = origin

    def __repr__(self) -> str:
        return (
            f'{self.__class__.__name__}("{self.provider_type}", {self.contents}, {self.minimum_pack_format},'
            f" {self.chars_covered})"
        )

    def __str__(self) -> str:
        return f"{self.__class__.__name__}({len(self.chars_covered):,} chars)"

    @abc.abstractmethod
    def validate(self) -> None:
        """
        Specific tests for validating a provider.

        This function checks values to ensure that they're correct and in an acceptable range.

        If applicable, textures are checked to ensure all characters have a defined glyph,
        and that the dimensions are correct.

        Warnings are printed through :func:`mcfonts.logger.warning`.

        :returns: Nothing; problems are either raised or warned.
        """

    @abc.abstractmethod
    def print_info(self) -> None:
        """Print information about this provider."""
        print(f"Provider type: {mcfonts.constants.UNKNOWN_FIELD}")
        print(f"Characters covered: {len(self.chars_covered):,}")

    @abc.abstractmethod
    def pretty_print(self) -> str:
        """
        Return a short one line description of the provider.

        :returns: A string of the provider's info, normally ``<icon> <type>: <info,...>``.
        """
        return f"{self.icon} {self.provider_type}"

    @abc.abstractmethod
    def construct_glyphs(
        self, ignore_empty: bool = True
    ) -> collections.abc.Iterable[tuple[str, mcfonts.glyphs.AnyGlyph]]:
        """
        Yield a tuple of (character, glyph) for every glyph available in the provider's resource.

        :param ignore_empty: Whether to ignore or include None for codepoints that don't have glyphs.
        :returns: A yield of ``(character, AnyGlyph)``.
        """

    @abc.abstractmethod
    def yield_characters(self, ignore_padding: bool = True) -> collections.abc.Iterable[str]:
        """
        Yield strings of the individual characters this provider supports.

        :returns: A yield of strings of length 1.
        """

    @abc.abstractmethod
    def yield_glyphs(self) -> collections.abc.Iterable[tuple[str, mcfonts.glyphs.AnyGlyph]]:
        """
        Yield a tuple of (character, glyph) for every glyph available in the provider's resource.

        This is different from :meth:`construct_glyphs`
        because it yields glyphs that have already been constructed and stored.

        :returns: A yield of ``(character, AnyGlyph)``.
        """



#!/usr/bin/env python3
# coding=utf-8
# SPDX-License-Identifier: MIT
"""A unihex provider, handling .hex glyphs."""
import collections.abc
import os

import jsonschema

import mcfonts.glyphs
import mcfonts.providers.base
import mcfonts.utils.schemas
import mcfonts.utils.unihex


class UnihexProvider(mcfonts.providers.base.BaseProvider):
    """
    The ``unihex`` provider loads glyphs from text files in the ".hex" format.

    It has two required fields, ``hex_file`` and ``size_overrides``.

    * ``hex_file`` is a resource location to a .ZIP file containing any number of ".hex" files at its root.
    * ``size_overrides`` is a list of maps stating the beginning and end of a character range, and the starting and
      ending columns of the glyphs that match that range.

    Each ".hex" file is a text document with each line containing a hex glyph.
    A hex glyph is a string of text in the format of ``<codepoint>:<bit string>``.
    The bit string is a string of hexadecimal numbers.
    The numbers, when converted to binary, describe the glyph's pixels.
    A 1 bit is an "on" pixel, and a 0 bit is an "off" pixel.
    When exporting, the empty left and right columns are trimmed unless ``size_overrides`` explicitly overrides this.
    """

    minimum_pack_format = 15
    icon = "⬣"

    def __init__(self, provider: dict, origin: os.PathLike[str] | str, resources: dict[str, str] | None):
        self.contents = provider
        self.resources = resources
        self.size_overrides: dict[str, tuple[int, int]] = dict(self.yield_size_overrides())
        self.glyphs: dict[str, mcfonts.glyphs.UnihexGlyph] | None = dict(self.construct_glyphs())
        super().__init__("unihex", provider, origin, set(self.glyphs.keys()))
        self.validate()

    def construct_glyphs(
        self, ignore_empty: bool = True
    ) -> collections.abc.Iterable[tuple[str, mcfonts.glyphs.UnihexGlyph]] | None:
        if not self.resources:
            return
        for name, hex_contents in self.resources.items():
            for hex_glyph in hex_contents.splitlines():
                yield mcfonts.utils.unihex.get_codepoint(hex_glyph), mcfonts.glyphs.UnihexGlyph(
                    mcfonts.utils.unihex.get_bit_string(hex_glyph), None, None, None, None
                )

    def yield_size_overrides(self) -> collections.abc.Iterable[tuple[str, tuple[int, int]]]:
        """
        Yield all the size overrides for every character.

        Overrides are given as ``tuple[int, int]``.

        :returns: A yield of (character, (left, right))
        """
        override: dict
        for override in self.contents.get("size_overrides", {}):
            left: int = override.get("left", -1)
            right: int = override.get("right", -1)
            for codepoint in range(ord(override.get("from", 0)), ord(override.get("to", 0))):
                yield chr(codepoint), (left, right)

    def get_size_override(self, character: str) -> tuple[int, int] | None:
        """
        The size overrides are a tuple of two integers of the starting and ending columns to cut off when exporting.
        """
        return self.size_overrides.get(character)

    def validate(self) -> None:
        jsonschema.validate(self.contents, mcfonts.utils.schemas.SCHEMA_PROVIDER_UNIHEX)

    def pretty_print(self) -> str:
        return (
            f"{self.icon} unihex: {len(self.resources or []):,} .hex files, {len(self.size_overrides):,} size overrides"
        )

    def print_info(self) -> None:
        pass

    def yield_glyphs(self) -> collections.abc.Iterable[tuple[str, mcfonts.glyphs.UnihexGlyph]]:
        yield from self.glyphs.items()

    def yield_characters(self, ignore_padding: bool = True) -> collections.abc.Iterable[str]:
        for key in self.glyphs:
            yield key



#!/usr/bin/env python3
# coding=utf-8
# SPDX-License-Identifier: MIT
"""A filter is a class used to control what characters and providers are included or not included when exporting."""
from __future__ import annotations

import enum

import unirange

import mcfonts.providers.options
import mcfonts.providers.utils


@enum.unique
class OptionPolicy(enum.Enum):
    # TODO Deprecated, remove soon
    """How to handle special options attached to characters."""

    INCLUDE = 0
    """Include characters that have options declared."""
    DISREGARD_OPTIONS = 1
    """Disregard the options for a char, even if declared (but still include the char)."""
    INCLUDE_ONLY = 2
    """Only include characters that have any options declared."""
    EXCLUDE = 3
    """Do not include characters that have any options declared.."""


@enum.unique
class CharPolicy(enum.Enum):
    """How to handle filtered characters."""

    INCLUDE = 0
    """Include only the characters in this filter."""
    EXCLUDE = 1
    """Do not include any of the characters in this filter."""


@enum.unique
class ProviderPolicy(enum.Enum):
    """How to handle filtered providers."""

    INCLUDE = 0
    """Include only these providers."""
    EXCLUDE = 1
    """Do not include these providers."""


class FilterResult(enum.Enum):
    """The result of a filter query."""

    CONTINUE = 0
    """No special treatment needed; this character is not filtered."""
    SKIP_CHAR = 1
    """Skip this character."""
    SKIP_PROVIDER = 2
    """Skip this provider."""
    DISCARD_CHAR_OPTIONS = 3
    """Character is allowed, but special options must not be considered."""


class Filter:
    """
    A filter used in exporting.

    Can filter characters based on their codepoint, and if they have options.
    Can filter providers based on their type.
    """

    def __init__(
        self,
        filtered_chars: set[str] | str | None = None,
        filtered_providers: set[type[mcfonts.providers.utils.AnyProvider]] | None = None,
        char_policy: CharPolicy = CharPolicy.EXCLUDE,
        provider_policy: ProviderPolicy = ProviderPolicy.EXCLUDE,
        option_policy: OptionPolicy = OptionPolicy.INCLUDE,
    ) -> None:
        if filtered_chars is None:
            filtered_chars = set()
        elif isinstance(filtered_chars, str):
            filtered_chars = unirange.unirange_to_characters(filtered_chars)
        if filtered_providers is None:
            filtered_providers = set()
        self.filtered_chars: set[str] = filtered_chars
        self.char_policy: CharPolicy = char_policy
        self.filtered_providers: set[type[mcfonts.providers.utils.AnyProvider]] = filtered_providers
        self.provider_policy: ProviderPolicy = provider_policy
        self.option_policy: OptionPolicy = option_policy  # TODO Deprecated, remove soon

    def filter_char(self, character: str) -> bool:
        """
        Add `character` to the character filter.

        :param character: A string of 1 character.
        :returns: Boolean of success; True if character was added, False if it was already present.
        """
        if character in self.filtered_chars:
            return False
        self.filtered_chars.add(character)
        return True

    def unfilter_char(self, character: str) -> bool:
        """
        Remove `character` from the character filter.

        :param character: A string of 1 character.
        :returns: Boolean of success; True if character was removed, False if it wasn't present.
        """
        if character not in self.filtered_chars:
            return False
        self.filtered_chars.remove(character)
        return True

    def filter_provider(self, provider: type[mcfonts.providers.utils.AnyProvider]) -> bool:
        """
        Add `provider` to the provider filter.

        :param provider: A type of :data:`~mcfonts.providers.utils.AnyProvider`.
        :returns: Boolean of success; True if the provider was added, False if it was already present.
        """
        if provider in self.filtered_providers:
            return False
        self.filtered_providers.add(provider)
        return True

    def unfilter_provider(self, provider: type[mcfonts.providers.utils.AnyProvider]) -> bool:
        """
        Remove `provider` from the provider filter.

        :param provider: A type of :data:`~mcfonts.providers.utils.AnyProvider`.
        :returns: Boolean of success; True if provider was removed, False if it was never present.
        """
        if provider not in self.filtered_providers:
            return False
        self.filtered_providers.remove(provider)
        return True

    def set_char_policy(self, new_char_policy: CharPolicy) -> None:
        """
        Set the character policy for how to handle filtering characters.

        :param new_char_policy: A value from the :class:`CharPolicy` enum.
        """
        self.char_policy = new_char_policy

    def set_provider_policy(self, new_provider_policy: ProviderPolicy) -> None:
        """
        Set the provider policy for how to handle filtering providers.

        :param new_provider_policy: A value from the :class:`ProviderPolicy` enum.
        """
        self.provider_policy = new_provider_policy

    def set_option_policy(self, new_option_policy: OptionPolicy) -> None:
        """
        Set the option policy for how to handle filtering characters with options.

        :param new_option_policy: A value from the :class:`OptionPolicy` enum.
        """
        self.option_policy = new_option_policy  # TODO Deprecated, remove soon

    def should_ignore_char(
        self, character: str, options: mcfonts.providers.options.OptionsProvider | None = None
    ) -> FilterResult:
        """
        Determine how `character` should be handled, working off of the current Filter instance.

        - If filtered characters are declared:
            - If character is **IN** filtered characters **AND** character policy is EXCLUDE
            - OR
            - If character is **NOT IN** filtered characters **AND** character policy is INCLUDE:
                - *Skip character*
        - If options exist:
            - If options **DO EXIST** for character:
                - If option policy is INCLUDE or INCLUDE_ONLY:
                    - *Continue*
                - If option policy is DISREGARD_OPTIONS:
                    - *Ignore character options, continue*
                - If option policy is EXCLUDE:
                    - *Skip character*
            - If options **DO EXIST** for character **AND** option policy is INCLUDE_ONLY:
                - *Skip character*
        - Else:
            - *Continue*

        :param character: A string of a character to determine filtering for.
        :param options:
            An optional instance of :class:`~mcfonts.providers.options.OptionsProvider`.
            If this is None, :data:`self.char_policy` will be ignored.
        :returns: A value from the :class:`FilterResult` enum.
        """

        if self.filtered_chars and (
            (self.char_policy == CharPolicy.EXCLUDE and character in self.filtered_chars)
            or (self.char_policy == CharPolicy.INCLUDE and character not in self.filtered_chars)
        ):
            return FilterResult.SKIP_CHAR

        if options is not None:
            if options.has_options(character):
                if self.option_policy in {
                    OptionPolicy.INCLUDE,
                    OptionPolicy.INCLUDE_ONLY,
                }:
                    return FilterResult.CONTINUE
                if self.option_policy == OptionPolicy.DISREGARD_OPTIONS:
                    return FilterResult.DISCARD_CHAR_OPTIONS
                if self.option_policy == OptionPolicy.EXCLUDE:
                    return FilterResult.SKIP_CHAR
            if self.option_policy == OptionPolicy.INCLUDE_ONLY:
                return FilterResult.SKIP_CHAR
        return FilterResult.CONTINUE

    def should_ignore_provider(self, provider: type[mcfonts.providers.utils.AnyProvider]) -> FilterResult:
        """
        Determine how `provider` should be handled, working off of the current Filter instance.

        - If filtered providers are declared:
            - If provider is **IN** filtered providers **AND** provider policy is EXCLUDE:
            - OR
            - If provider is **NOT IN** filtered providers **AND** provider policy is INCLUDE:
                - *Skip provider*
        - Else:
            - *Continue*

        :param provider: The type of the provider to determine filtering for; this is not an instance.
        :returns: A value from the :class:`FilterResult` enum.
        """
        if self.filtered_providers and (
            (provider in self.filtered_providers and self.provider_policy == ProviderPolicy.EXCLUDE)
            or (provider not in self.filtered_providers and self.provider_policy == ProviderPolicy.INCLUDE)
        ):
            return FilterResult.SKIP_PROVIDER
        return FilterResult.CONTINUE



#!/usr/bin/env python3
# coding=utf-8
# SPDX-License-Identifier: MIT
"""A reference provider, including other fonts."""
import collections.abc
import os
import typing

import jsonschema

import mcfonts.exceptions
import mcfonts.glyphs
import mcfonts.providers.base
import mcfonts.providers.bitmap
import mcfonts.providers.legacy_unicode
import mcfonts.providers.space
import mcfonts.providers.ttf
import mcfonts.providers.unihex
import mcfonts.providers.utils
import mcfonts.utils.schemas

AnyProviderNotReference = (
    mcfonts.providers.bitmap.BitmapProvider
    | mcfonts.providers.space.SpaceProvider
    | mcfonts.providers.legacy_unicode.LegacyUnicodeProvider
    | mcfonts.providers.ttf.TTFProvider
    | mcfonts.providers.unihex.UnihexProvider
)
"""
A UnionType similar to :class:`mcfonts.providers.utils.AnyProvider`,
except that it does not include :class:`ReferenceProvider`."""

PROVIDER_SHORTCUT_MAP: dict[typing.Type[AnyProviderNotReference], str] = {
    mcfonts.providers.bitmap.BitmapProvider: "bt",
    mcfonts.providers.space.SpaceProvider: "sp",
    mcfonts.providers.legacy_unicode.LegacyUnicodeProvider: "lu",
    mcfonts.providers.ttf.TTFProvider: "ttf",
    mcfonts.providers.unihex.UnihexProvider: "uh",
}


class ReferenceProvider(mcfonts.providers.base.BaseProvider):
    """
    The ``reference`` provider is used to include and load another font only once.

    It has one field, ``id``, which points to a font JSON to include.
    If more than 1 reference provider points to the same font JSON, it will be ignored.
    """

    minimum_pack_format = 15
    icon = "↩"

    def __init__(self, provider: dict, origin: os.PathLike[str] | str, children: list[AnyProviderNotReference]):
        if isinstance(children, self.__class__):
            raise mcfonts.exceptions.ReferenceChildError("a child provider cannot be another reference provider")
        self.children: list[AnyProviderNotReference] = children

        super().__init__("reference", provider, origin, set(self.yield_characters()))
        self.validate()

    def validate(self) -> None:
        jsonschema.validate(self.contents, mcfonts.utils.schemas.SCHEMA_PROVIDER_REFERENCE)

    def print_info(self) -> None:
        super().print_info()
        n_children: int = len(self.children)
        print(f"Children: ({n_children})")
        provider: mcfonts.providers.utils.AnyProvider
        for provider in self.children:
            print(f"\t- {str(provider.__class__)}, {len(provider.chars_covered):,} chars")

    def construct_glyphs(
        self, ignore_empty: bool = True
    ) -> collections.abc.Iterable[tuple[str, mcfonts.glyphs.AnyGlyph]]:
        # TODO add this stuff I guess
        yield from []

    def yield_children(self) -> collections.abc.Iterable[AnyProviderNotReference]:
        """
        Yield all the providers that this reference provider is referring to.

        :returns: A yield of the contained providers.
        """
        yield from self.children

    def pretty_print(self) -> str:
        children_formatted: list[str] = [
            PROVIDER_SHORTCUT_MAP.get(child.__class__, mcfonts.constants.UNKNOWN_FIELD) for child in self.children
        ]
        return f"{self.icon} reference: {len(self.children):,} children: [{', '.join(children_formatted)}]"

    def yield_glyphs(self) -> collections.abc.Iterable[mcfonts.glyphs.BitmapGlyph | None]:
        raise NotImplementedError()

    def yield_characters(self, ignore_padding: bool = True) -> collections.abc.Iterable[str]:
        for child in self.children:
            for char in child.yield_characters(ignore_padding):
                yield char



#!/usr/bin/env python3
# coding=utf-8
# SPDX-License-Identifier: MIT
"""A legacy unicode provider, handling bitmap glyphs in a 16px x 16px grid."""
import collections.abc
import os
import re
import unicodedata

import PIL.Image
import jsonschema

import mcfonts.constants
import mcfonts.glyphs
import mcfonts.providers.base
import mcfonts.utils.resources
import mcfonts.utils.schemas


class LegacyUnicodeProvider(mcfonts.providers.base.BaseProvider):
    """
    The ``legacy_unicode`` provider is a "fallback" provider intended to be used as a last-resort.

    It is similar to the :class:`bitmap provider<mcfonts.providers.bitmap.BitmapProvider>`
    in that its glyphs are bitmaps.

    It uses a system of templates to create and add 16px x 16px tables of "fallback" characters.
    Each glyph is 16px x 16px pixels wide, so each page is 256px x 256px;
    16 characters on each line, 16 lines, 16 length for each glyph.

    .. warning::
        This provider is deprecated and should not be used when possible.
        Use the "unihex" provider instead.

    .. important::
        Characters above the `BMP <https://en.wikipedia.org/wiki/Plane_(Unicode)#Basic_Multilingual_Plane>`_ (U+FFFF)
        are not handled.
    """

    minimum_pack_format = 4
    icon = "🖳"

    def __init__(self, provider: dict, origin: os.PathLike[str] | str, resources: dict[str, PIL.Image.Image | bytes]):
        self.resources = resources
        self.codepages_covered: list[int] = []
        self.glyphs: dict[str, mcfonts.glyphs.BitmapGlyph] = dict(self.construct_glyphs())
        super().__init__(
            "legacy_unicode",
            provider,
            origin,
            set(self.yield_characters()),
        )
        self.validate()

    def pretty_print(self) -> str:
        return f"{self.icon} legacy_unicode: {self.contents.get('template', 'no template')}"

    def validate(self) -> None:
        jsonschema.validate(self.contents, mcfonts.utils.schemas.SCHEMA_PROVIDER_LEGACY_UNICODE)

    def print_info(self) -> None:
        # TODO add the format here
        """Print information about this provider."""
        print(f"Template: {self.contents['template'] or mcfonts.constants.UNKNOWN_FIELD}")
        print(f"Sizes: {self.contents['sizes'] or mcfonts.constants.UNKNOWN_FIELD}")

    def construct_glyphs(
        self, ignore_empty: bool = True
    ) -> collections.abc.Iterable[tuple[str, mcfonts.glyphs.AnyGlyph]]:
        yield from []

    def yield_glyphs(
        self, ignore_empty: bool = True
    ) -> collections.abc.Iterable[tuple[str, mcfonts.glyphs.BitmapGlyph]]:
        for codepage, content in self.resources.items():
            # If template_\d\d\.png doesn't match codepage name, skip.
            if not re.match(
                self.contents["template"].replace("\\", "\\\\").replace("%s", r"(\d\d)"), codepage
            ) or not isinstance(content, PIL.Image.Image):
                continue
            codepage_num = re.match(codepage, r"\d\d")
            for index, glyph in enumerate(mcfonts.utils.resources.divide_resource_by_grid(content, (16, 16))):
                if glyph is None:
                    continue
                yield chr((int(codepage_num.group(1), 16) << 8) + index), mcfonts.glyphs.BitmapGlyph(glyph, 1)

    def yield_characters(self, ignore_padding: bool = True) -> collections.abc.Iterable[str]:
        for key in self.glyphs.keys():
            yield str(key)


def to_advances(glyphsizes: bytes, match_unicode_category: list[str] | None = None) -> dict[str, int]:
    """
    Translate a glyphsizes.bin file into an "advances" mapping, which goes inside a "space" provider.

    .. info::

        This function does not return a new provider.

    >>> to_advances(b"\\x00\\x00\\x00...")
    {
        "a": 0,
        "b": 0,
        "c": 0,
        "\uBEEF": 15
    }

    :param glyphsizes: The bytes of glyphsizes.bin.
    :param match_unicode_category:
        Only translate Unicode characters with these categories.
        By default, this is ``[Mc, Zp, Zs, Zl]``.
        This should cover most whitespace and marking characters.
    :returns: An "advances" dictionary.
    """
    if match_unicode_category is None:
        match_unicode_category = ["Mc", "Zp", "Zs", "Zl"]
    advances = {}
    for index, nibble in enumerate(glyphsizes):
        if unicodedata.category(char := chr(index)) in match_unicode_category:
            advances[char] = (nibble & 0xF) - (nibble >> 4 & 0xF)
    return advances


def build_template_for_char(char: str, path_to_template_textures: str | None = None) -> PIL.Image.Image:
    """
    Build an image of a template :term:`codepoint`.

    Templates are simple white boxes with a codepoint :term:`texture` inside them.
    This is intended for use with a "legacy_unicode" :term:`provider`.

    :param char:
        A single character.
        This character's codepoint is what goes inside the box.
    :param path_to_template_textures:
        The path to the folder that contains the digit textures.
        See the documentation for :func:`mcfonts.templates.generate_unicode_template_font`.
    :returns: A :class:`PIL.Image.Image` of the char template
    """
    codepoint = ord(char)
    templates = generate_unicode_template_font(path_to_template_textures)
    if codepoint > 0xFFFF:
        box = templates[-6]
        # 0xABCDEF -> A
        box.paste(templates[(codepoint & 0xF00000) >> 20], (2, 2))
        # 0xABCDEF -> B
        box.paste(templates[(codepoint & 0xF0000) >> 16], (6, 2))
        # 0xABCDEF -> C
        box.paste(templates[(codepoint & 0xF000) >> 12], (10, 2))
        # 0xABCDEF -> D
        box.paste(templates[(codepoint & 0xF00) >> 8], (2, 9))
        # 0xABCDEF -> E
        box.paste(templates[(codepoint & 0xF0) >> 4], (6, 9))
        # 0xABCDEF -> F
        box.paste(templates[(codepoint & 0xF)], (10, 9))
    else:
        box = templates[-4]
        # 0xABCD -> A
        box.paste(templates[(codepoint & 0xF000) >> 12], (2, 2))
        # 0xABCD -> B
        box.paste(templates[(codepoint & 0xF00) >> 8], (6, 2))
        # 0xABCD -> C
        box.paste(templates[(codepoint & 0xF0) >> 4], (2, 9))
        # 0xABCD -> D
        box.paste(templates[codepoint & 0xF], (6, 9))
    return box


def generate_unicode_template_font(path: str | None = None) -> dict[int, PIL.Image.Image]:
    """
    Generate a dictionary of hex -> sheet.

    Path must point to a folder where these files exist (PNG format):

    * ``<0-9>.png``
    * ``<a-f>.png``
    * ``box4.png``
    * ``box6.png``

    The file names **are** case-sensitive, and the textures for hex charlist **must** be 3 pixels wide by 6 pixels tall.

    * The sheet for "box6.png" must be 16px x 16px.
    * The sheet for "box4.png" must be 12px x 16px.
    * Hex characters are overlayed atop the box.
    * Box4 is for chars below U+10000 (4 points, F4E -> 0F4E; 4).
    * Box6 is for chars above U+FFFF (6 points, 1FAE6 -> 01FAE6; 6).

    Color does not matter.

    This is used for generating Unicode template characters,
    see :func:`mcfonts.utils.legacy_unicode.build_template_for_char`.

    :param path:
        Path to the folder where these textures exist.
        If this is falsy (None, ""), the path will be set automatically from the module and will
        use pre-made character textures that were distributed with this module.
    :returns:
        A dictionary that maps 0-15 to a glyph.
        -4 and -6 are box4 and box6, respectively.
    """
    if path:
        templates = mcfonts.utils.resources.expand_path(path)
    else:
        templates = mcfonts.utils.resources.expand_path(os.path.join(__file__, "../../template_hexchar_textures"))
    return {
        0: PIL.Image.open(f"{templates}/0.png"),
        1: PIL.Image.open(f"{templates}/1.png"),
        2: PIL.Image.open(f"{templates}/2.png"),
        3: PIL.Image.open(f"{templates}/3.png"),
        4: PIL.Image.open(f"{templates}/4.png"),
        5: PIL.Image.open(f"{templates}/5.png"),
        6: PIL.Image.open(f"{templates}/6.png"),
        7: PIL.Image.open(f"{templates}/7.png"),
        8: PIL.Image.open(f"{templates}/8.png"),
        9: PIL.Image.open(f"{templates}/9.png"),
        10: PIL.Image.open(f"{templates}/a.png"),
        11: PIL.Image.open(f"{templates}/b.png"),
        12: PIL.Image.open(f"{templates}/c.png"),
        13: PIL.Image.open(f"{templates}/d.png"),
        14: PIL.Image.open(f"{templates}/e.png"),
        15: PIL.Image.open(f"{templates}/f.png"),
        -4: PIL.Image.open(f"{templates}/box4.png"),
        -6: PIL.Image.open(f"{templates}/box6.png"),
    }


def align_unicode_page(sheet: PIL.Image.Image) -> PIL.Image.Image:
    """
    Align a Unicode page font sheet's characters to the left.

    This function is a shortcut for :func:`mcfonts.utils.resources.align_font_texture(sheet, (16, 16))`.

    :param sheet: The font sheet, not the individual character.
    :returns: The new font sheet.
    """
    return mcfonts.utils.resources.align_font_texture(sheet, (16, 16))



#!/usr/bin/env python3
# coding=utf-8
# SPDX-License-Identifier: MIT
"""
The Options provider.

.. warning::
    This is deprecated and will be removed in v0.7.
"""
import os
import warnings

import jsonschema
import unirange

import mcfonts.utils.schemas

ACCEPTED_OPTIONS_TYPES: set[str] = {"mcfonts:options", "options"}

warnings.warn("The options provider is deprecated.", DeprecationWarning, stacklevel=2)


class OptionsProvider:
    """
    A mcfonts-specific provider that only is used when exporting.

    This does not extend :class:`mcfonts.providers.base.BaseProvider`.

    .. warning::
        This is deprecated and will be removed in v0.7.
    """

    # TODO Deprecated, removing in v0.7.

    icon = "⚙"

    def __init__(self, provider: dict, origin: os.PathLike[str] | str):
        warnings.warn("The options provider is deprecated.", DeprecationWarning, stacklevel=2)
        self.contents = provider
        self.origin = origin
        self.contents["fields"] = {}

        # Expand the uniranges, copy over fields
        for field in OPTION_FIELDS:
            if field not in self.contents["fields"]:
                self.contents["fields"][field] = {}
            for key, value in provider.get("fields", {}).get(field, {}).items():
                for newchar in unirange.unirange_to_characters(key):
                    self.contents["fields"][field][newchar] = value
        self.chars_covered: set[str] = set()
        for field in OPTION_FIELDS:
            self.chars_covered.update(self.contents["fields"].get(field, {}).keys())
        self.validate()

    def validate(self) -> None:
        """
        Validate this provider, ensuring its structure is correct and that all values are acceptable.

        This function **does** validate textures.
        """
        jsonschema.validate(self.contents, mcfonts.utils.schemas.SCHEMA_PROVIDER_OPTIONS)
        for char in self.chars_covered:
            if self.contents.get("fullwidth", {}).get(char) and self.contents.get("spacing", {}).get(char):
                mcfonts.logger.warning(
                    f"Provider has conflicting options, fullwidth and spacing are present for character {char}."
                )
            if self.contents.get("fullwidth", {}).get(char) and self.contents.get("width", {}).get(char):
                mcfonts.logger.warning(
                    f"Provider has conflicting options, fullwidth and width are present for character {char}."
                )
            if self.contents.get("width", {}).get(char) and self.contents.get("spacing", {}).get(char):
                mcfonts.logger.warning(
                    f"Provider has conflicting options, width and spacing are present for character {char}."
                )

    def get_options_for_char(self, char: str) -> dict[str, int | bool | list[int]] | None:
        """
        Given `char`, a character, return a dictionary of all the options for that character.

        For example, if :attr:`self.contents` is:
            {"width": {"a": 4}, "spacing": {"a": 0}}
        Then `self.get_options_for_char("a")` should return:
            {"width": 4, "spacing": 0}

        :param char: A single character.
        :returns: A flat dictionary containing a value for each field that `char` has declared.
        """
        if char in mcfonts.constants.PADDING_CHARS:
            return None
        result = {}
        for field in OPTION_FIELDS:
            if (char_options := self.contents["fields"].get(field, {}).get(char)) is not None:
                result[field] = char_options
        if result:
            return result
        return None

    def has_options(self, char: str, field_names: str | set[str] | None = None) -> bool:
        """
        Given `char`, a character, return if any options are defined for it.

        If `field_names` is not None, only return True if the character has an option for that field.

        :param char:
            A single character.
        :param field_names:
            Either a string, list of strings, or None.
            * If str, check for this field name only.
            * If list[str], check for *any* of these field names.
            * If None, no special field check.
        :returns: If `char` has an option declared for it.
        """
        if isinstance(field_names, str):
            fields = {field_names}
        elif isinstance(field_names, set):
            fields = field_names
        else:
            fields = OPTION_FIELDS
        for field in fields:
            if (self.contents["fields"].get(field, {}).get(char)) is not None:
                return True
        return False

    def pretty_print(self) -> str:
        """Pretty print this provider."""
        return f"{self.icon} options: {len(self.chars_covered):,} chars"


OPTION_FIELDS: set[str] = {
    "width",
    "fullwidth",
    "shift",
    "spacing",
}



#!/usr/bin/env python3
# coding=utf-8
# SPDX-License-Identifier: MIT
"""Utilities and types for the providers."""
from __future__ import annotations

import collections.abc
import json
import os

import PIL

import mcfonts.exceptions
import mcfonts.providers.bitmap
import mcfonts.providers.legacy_unicode
import mcfonts.providers.options
import mcfonts.providers.reference
import mcfonts.providers.space
import mcfonts.providers.ttf
import mcfonts.providers.unihex
import mcfonts.utils.resources


AnyProvider = (
    mcfonts.providers.bitmap.BitmapProvider
    | mcfonts.providers.space.SpaceProvider
    | mcfonts.providers.legacy_unicode.LegacyUnicodeProvider
    | mcfonts.providers.ttf.TTFProvider
    | mcfonts.providers.unihex.UnihexProvider
    | mcfonts.providers.reference.ReferenceProvider
)
"""
A UnionType for any kind of Provider subclass that the Vanilla game accepts.
This does not include :class:`OptionsProvider`.
"""


def format_provider_message(
    provider: AnyProvider | mcfonts.providers.options.OptionsProvider | dict, message: str
) -> str:
    """
    Format a provider warning message properly, following Provider <type>: <message>.

    Calls :func:`mcfonts.utils.pretty_print_provider()`.

    :param provider:
        The provider, either as a dictionary or instance of :data:`mcfonts.providers.AnyProvider`.
    :param message: The message to append at the end.
    :returns: The formatted warning message.
    """
    if isinstance(provider, dict):
        return f"Provider {pretty_print_provider_dictionary(provider)} {message}"
    return f"Provider {provider.pretty_print()} {message}"


def pretty_print_provider_dictionary(provider: dict) -> str:
    """
    Format a provider information message properly, following ``<type><provider specific>``.

    Provider specific info:

     * ``bitmap``: ``<file> h <height> a <ascent>``
     * ``space``: nothing
     * ``legacy_unicode``: ``<template>``
     * ``ttf``: ``<file> s <shift0, 1>, sz <size>, o <oversample>, sk <skip>``

    :param provider:
        The provider as a dictionary, **not** an instance of :data:`mcfonts.providers.AnyProvider`.
    :returns: The pretty provider information.
    """
    if (provider_type := provider.get("type", "")).lower() == "bitmap":
        return f'"bitmap": {provider.get("file", "no resource")}'
    if provider_type == "space":
        return '"space"'
    if provider_type == "legacy_unicode":
        return f'"legacy_unicode": {provider.get("template", "no template")}'
    if provider_type == "ttf":
        return (
            f'"ttf": {provider.get("file", "no file")}, s '
            f'{provider.get("shift", "[?, ?]")}, sz '
            f'{provider.get("size", "?")}, o '
            f'{provider.get("oversample", "?")}, sk '
            f'{provider.get("skip", "none")}'
        )
    if provider_type == "options":
        return '"options"'
    return f'provider "{provider_type}": ?'


def load_all(
    providers: list[dict],
    json_path: os.PathLike[str] | str,
    strict: bool = False,
    mode: str = "LA",
) -> collections.abc.Iterable[mcfonts.providers.utils.AnyProvider | mcfonts.providers.options.OptionsProvider]:
    """
    Load every provider in the list `providers`.

    :param providers:
        A list of dictionaries of providers.
        This should be the value of the "providers" key in the JSON file.
    :param json_path: The path to the JSON file at which this provider JSON is from.
    :param strict: Whether to raise or warn (and continue) on exceptions.
    :param mode:
        What "mode" to load bitmap resources in.

        * ``"LA"`` - Grayscale
        * ``"RGBA"`` - Color

    :returns: A yield of AnyProvider | OptionsProvider.
    """
    # TODO on deprecation, remove | OptionsProvider.
    for provider in providers:
        yield load_provider(provider, json_path, strict, mode)


def load_provider(
    provider: dict,
    json_path: os.PathLike[str] | str,
    strict: bool = False,
    mode: str = "LA",
) -> mcfonts.providers.utils.AnyProvider | mcfonts.providers.options.OptionsProvider:
    """
    Load a specific provider dictionary and return the equivalent Provider class.

    :param provider: A dictionary of a provider.
    :param json_path: The path to the JSON file at which this provider JSON is from.
    :param strict: Whether to raise or warn (and continue) on exceptions.
    :param mode:
        What "mode" to load bitmap resources in.

        * ``"LA"`` - Grayscale
        * ``"RGBA"`` - Color

    :returns: A Provider instance.
    """
    provider_type = provider.get("type", "").lower().strip()
    try:
        if provider_type == "bitmap":
            return mcfonts.providers.bitmap.BitmapProvider(
                provider,
                json_path,
                mcfonts.utils.resources.load_bitmap(
                    provider["file"], mcfonts.utils.resources.traverse_to_assets(json_path)
                ),
            )
        elif provider_type == "legacy_unicode":
            return mcfonts.providers.legacy_unicode.LegacyUnicodeProvider(
                provider,
                json_path,
                dict(
                    mcfonts.utils.resources.load_resources_legacy_unicode(
                        provider["template"], mcfonts.utils.resources.traverse_to_assets(json_path), provider["sizes"]
                    )
                ),
            )
        elif provider_type == "space":
            return mcfonts.providers.space.SpaceProvider(provider, json_path)
        elif provider_type == "ttf":
            with open(
                mcfonts.utils.resources.resolve_path(
                    provider["file"], mcfonts.utils.resources.traverse_to_assets(json_path), "font"
                ),
                "rb",
            ) as open_tempfile:
                ttf = open_tempfile.read()
            return mcfonts.providers.ttf.TTFProvider(provider, json_path, ttf)
        elif provider_type == "unihex":
            return mcfonts.providers.unihex.UnihexProvider(
                provider,
                json_path,
                dict(
                    mcfonts.utils.resources.load_resources_unihex(
                        provider["hex_file"], mcfonts.utils.resources.traverse_to_assets(json_path)
                    )
                ),
            )
        elif provider_type == "reference":
            # Create any child providers first
            with open(
                temp_recursive_path := mcfonts.utils.resources.resolve_path(
                    provider["id"] + ".json", mcfonts.utils.resources.traverse_to_assets(json_path), "font"
                ),
                encoding="utf-8",
            ) as temp_recursive_file:
                data = json.load(temp_recursive_file, strict=strict)
            # This is recursive
            children = list(load_all(data.get("providers", {}), temp_recursive_path, strict, mode))
            if any(isinstance(child, mcfonts.providers.reference.ReferenceProvider) for child in children):
                raise mcfonts.exceptions.ReferenceChildError()
            return mcfonts.providers.reference.ReferenceProvider(provider, json_path, children)
        elif provider_type in mcfonts.providers.options.ACCEPTED_OPTIONS_TYPES:
            return mcfonts.providers.options.OptionsProvider(provider, json_path)
    except PIL.UnidentifiedImageError as exception:
        if not strict:
            mcfonts.logger.warning(
                mcfonts.providers.utils.format_provider_message(provider, "has invalid file; must be PNG.")
            )
        else:
            raise exception
    except FileNotFoundError as exception:
        if not strict:
            # Missing files are ignored
            mcfonts.logger.warning(
                mcfonts.providers.utils.format_provider_message(
                    provider, f'file "{exception.filename}" does not exist, skipping.'
                )
            )
        else:
            # Missing files not allowed, raising exception
            raise exception
    raise mcfonts.exceptions.ProviderError(f'Unknown provider "{provider_type}".')



#!/usr/bin/env python3
# coding=utf-8
# SPDX-License-Identifier: MIT
"""A ttf provider, including already-made TTF and OTF fonts."""
import collections.abc
import os

import jsonschema

import mcfonts.glyphs
import mcfonts.providers.base
import mcfonts.utils.schemas


class TTFProvider(mcfonts.providers.base.BaseProvider):
    """
    The ``ttf`` provider embeds already-compiled TrueType and OpenType fonts.

    mcfonts does not handle TTF providers by default, and they cannot be :doc:`exported <exporting>`.
    """

    minimum_pack_format = 9
    icon = "𝑇"

    def __init__(self, provider: dict, origin: os.PathLike[str] | str, resource: bytes | None = None):
        super().__init__("ttf", provider, origin)
        if "skip" in self.contents:
            chars = []
            if isinstance(skip := self.contents.get("skip"), list):
                for line in skip:
                    chars.append(line)
            # Flatten into a string
            self.contents["skip"] = "".join(chars)
        self.validate()
        self.resource = resource

    def validate(self) -> None:
        jsonschema.validate(self.contents, mcfonts.utils.schemas.SCHEMA_PROVIDER_TTF)

    def pretty_print(self) -> str:
        return f"{self.icon} ttf: {self.contents.get('file', 'no file')}"

    def print_info(self) -> None:
        super().print_info()
        # TODO finish thsi

    def construct_glyphs(
        self, ignore_empty: bool = True
    ) -> collections.abc.Iterable[tuple[str, mcfonts.glyphs.AnyGlyph]]:
        yield from []

    def yield_glyphs(self, ignore_empty: bool = True) -> collections.abc.Iterable[tuple[str, mcfonts.glyphs.AnyGlyph]]:
        yield from []

    def yield_characters(self, ignore_padding: bool = True) -> collections.abc.Iterable[str]:
        yield from []



#!/usr/bin/env python3
# coding=utf-8
# SPDX-License-Identifier: MIT
"""JSON schemas for providers."""
from __future__ import annotations

import re

import unirange

PATTERN_RESOURCE_LOCATION = re.compile(r"^([0-9a-z_.]+:)?[0-9a-z_./-]+")
PATTERN_RESOURCE_LOCATION_TEMPLATED = re.compile(r"^([0-9a-z_.]+:)?[0-9a-z_./%-]+")

SCHEMA_PROVIDER_BITMAP = {
    "type": "object",
    "properties": {
        "type": {"const": "bitmap"},
        "file": {"type": "string", "minLength": 1, "pattern": PATTERN_RESOURCE_LOCATION.pattern + r"\.png$"},
        "height": {"type": "integer", "minimum": 1, "maximum": 512, "default": 8},
        "ascent": {"type": "integer", "minimum": -512, "maximum": 512},
        "chars": {"type": "array", "minItems": 1, "items": {"type": "string"}},
        "__comment": {"type": "string"},
    },
    "required": ["type", "file", "ascent", "chars"],
    "additionalProperties": True,
}
"""A schema to validate "bitmap" providers against."""

SCHEMA_PROVIDER_SPACE = {
    "type": "object",
    "properties": {
        "type": {"const": "space"},
        "advances": {
            "type": "object",
            "minLength": 1,
            "patternProperties": {unirange.UNIRANGE_FULL.pattern: {"type": "integer", "minimum": -256, "maximum": 256}},
        },
        "comment": {"type": "string"},
    },
    "required": ["type", "advances"],
    "additionalProperties": True,
}

"""A schema to validate "space" providers against."""

SCHEMA_PROVIDER_LEGACY_UNICODE = {
    "type": "object",
    "properties": {
        "type": {"const": "legacy_unicode"},
        "sizes": {"type": "string", "minLength": 1, "pattern": PATTERN_RESOURCE_LOCATION.pattern + r"\.bin$"},
        "template": {
            "type": "string",
            "minLength": 1,
            "pattern": PATTERN_RESOURCE_LOCATION_TEMPLATED.pattern + r"\.png$",
        },
        "comment": {"type": "string"},
    },
    "required": ["type", "sizes", "template"],
    "additionalProperties": True,
}
"""A schema to validate "legacy_unicode" providers against."""

SCHEMA_PROVIDER_TTF = {
    "type": "object",
    "properties": {
        "type": {"const": "ttf"},
        "file": {"type": "string", "minLength": 1, "pattern": PATTERN_RESOURCE_LOCATION.pattern + r"\.[ot]tf$"},
        "shift": {"type": "array", "items": {"type": "number"}, "minItems": 2, "maxItems": 2},
        "size": {"type": "number", "minimum": 1},
        "oversample": {"type": "number", "minimum": 1},
        "skip": {"type": ["string", "array"]},
        "comment": {"type": "string"},
    },
    "required": ["type", "file", "shift", "size", "oversample"],
    "additionalProperties": True,
}
"""A schema to validate "ttf" providers against."""

SCHEMA_PROVIDER_OPTIONS = {
    "type": "object",
    "properties": {
        "type": {"enum": ["options", "mcfonts:options"]},
        "fields": {
            "type": "object",
            "properties": {
                "fullwidth": {
                    "type": "object",
                    "patternProperties": {unirange.UNIRANGE_FULL.pattern: {"type": "boolean"}},
                },
                "shift": {
                    "type": "object",
                    "patternProperties": {
                        unirange.UNIRANGE_FULL.pattern: {
                            "type": "array",
                            "minItems": 2,
                            "maxItems": 2,
                            "items": {"type": "integer"},
                        }
                    },
                },
                "spacing": {
                    "type": "object",
                    "patternProperties": {unirange.UNIRANGE_FULL.pattern: {"type": "integer", "minimum": 0}},
                },
                "width": {
                    "type": "object",
                    "patternProperties": {unirange.UNIRANGE_FULL.pattern: {"type": "integer"}},
                },
            },
        },
        "comment": {"type": "string"},
    },
    "required": ["type"],
    "additionalProperties": True,
}
"""A schema to validate "options" providers against."""

SCHEMA_PROVIDER_UNIHEX = {
    "type": "object",
    "properties": {
        "type": {"const": "unihex"},
        "hex_file": {"type": "string", "minLength": 1, "pattern": PATTERN_RESOURCE_LOCATION.pattern + r"\.zip$"},
        "size_overrides": {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {
                    "__ranges": {
                        # "__ranges" isn't part of the specification, but it's included in Vanilla files anyway.
                        "type": "array",
                        "items": {"type": "string", "minLength": 1},
                    },
                    "from": {"type": "string", "minLength": 1, "maxLength": 1},
                    "to": {"type": "string", "minLength": 1, "maxLength": 1},
                    "left": {"type": "integer", "minimum": 0, "maximum": 256},
                    "right": {"type": "integer", "minimum": 0, "maximum": 256},
                },
                "required": ["from", "to", "left", "right"],
                "additionalProperties": False,
            },
        },
    },
    "required": ["type", "hex_file", "size_overrides"],
    "additionalProperties": True,
}

SCHEMA_PROVIDER_REFERENCE = {
    "type": "object",
    "properties": {
        "type": {"const": "reference"},
        "id": {"type": "string", "minLength": 1, "pattern": PATTERN_RESOURCE_LOCATION.pattern},
    },
    "required": ["type", "id"],
    "additionalProperties": True,
}
